/* see W&Z 2000 Eq.(6) for the approximation */

#include <math.h>
#include <gsl/gsl_sf_bessel.h>
#include "constants.h"
#include "decs.h"

static double K2_WZ_high_T_eq(double tdl);
static double K2_WZ_low_T_eq(double tdl);

/*============================================================================*/
/* K2(1/tdl) for high tdl
 * see W&Z 2000 Eq.(4) */
static double K2_WZ_high_T_eq(double tdl)
{
        double T2 = tdl*tdl;
        double log2T = log(2.*tdl);
        return( 2*T2
               - 0.5
               + (log2T+0.75-C_euler)/(8.*T2)
               + (log2T+0.95)/(96.*T2*T2) );
}

/*----------------------------------------------------------------------------*/
/* K2(1/tdl) exp(1/tdl) for low tdl
 * see W&Z 2000 Eq.(4) */
static double K2exp_WZ_low_T_eq(double tdl)
{
        return( sqrt(C_pi_2*tdl)*(1.+15.*tdl/8.+105.*tdl*tdl/128.-0.203*tdl*tdl*tdl) );
}

/*============================================================================*/
/* Calculate 1/K_2(1/tdl) or  1/(K_2(1/tdl)exp(1/tdl))  */
double K2_tdlinv_inv(double tdlinv)
{
        double ans;
#if K2_EQ == K2_WZ_whole_eq
        double tdl = 1./tdlinv;
        if( tdl <= 0.65 ) {
                /* 1 / ( K_2(1/tdl)exp(1/tdl) )
                 * the factor 1 / exp(-1/tdl) is included in the integrand */
                ans = K2exp_WZ_low_T_eq(tdl);
        } else {
                /* 1 / ( K_2(1/tdl) ) */
                ans = K2_WZ_high_T_eq(tdl);
        }
        ans = 1./ans;
#elif K2_EQ == K2_WZ_high_T
        /* 1 / ( K_2(1/tdl) ) */
        ans = K2_WZ_high_T_eq(1./tdlinv);
        ans = 1./ans;
#elif K2_EQ == K2_GSL
        /* 1 / ( K_2(1/tdl)exp(1/tdl) )
         * the factor 1 / exp(-1/tdl) is included in the integrand */
        int status;
        gsl_sf_result result;
        status = gsl_sf_bessel_Kn_scaled_e(2, tdlinv, &result);
        ans = 1./result.val;
#elif K2_EQ == K2_GSL_interpolate
  #error "ERROR: K2_EQ == K2_GSL_interpolate not yet implemented"
#else
  #error "K2_EQ not defined correctly"
#endif
        return(ans);
}

/*----------------------------------------------------------------------------*/
/* Calculate 1/K_2(1/tdl) */
double K2_tdlinv_inv_new(double tdlinv)
{
        double tdl = 1./tdlinv;
        double ans;
        if( tdl <= 0.65 ) {
                ans = exp_factor( K2exp_WZ_low_T_eq(tdl) , -tdlinv );
        } else {
                ans = K2_WZ_high_T_eq(tdl);
        }
        return(1./ans);
}

/*============================================================================*/
