#include "decs.h"

/*============================================================================*/

/* example : isotropic ultrarelativistic single-electron emissivity
 *
 * Synchrotron emissivity j_\nu, from frequency = 1 \nu_c to 10^4 \nu_c, 
 *  at 45 degree.
 *
 * Compare j_\nu to ultrarelativistic solution in Rybicki & Lightman Eq.(6.33)
 *  (with K_5/3 and integration done by using gsl)
 *
 * Numerical calculation is done by approximating the Dirac Delta function by
 *  a narrow Gaussian function */

/* sanity check */
#if (ELECTRON_DIST != GAUSSIAN)
  #error This file assumes GAUSSIAN distribution is defined.  Set it in decs.h
#endif

/*============================================================================*/

double funcRL(double x, void *para)
{
  return(gsl_sf_bessel_Knu(5./3,x));
}

int main(int argc, char *argv[])
{
        double nrhoe = 1.e5;
        double B = 30.;
        /* temperature in K */
        double Te = 1.e10; /* unused in this case */
        /* observer angle */
        double theta = (45.*C_pi/180.);
        /* number of data point, each for one frequency */
        int npt=21;
        int i;

        /* allocation of memory for gamma-n integration: */
        gsl_integration_workspace *wksp_g = gsl_integration_workspace_alloc(N_WKSP);
        gsl_integration_workspace *wksp_n = gsl_integration_workspace_alloc(N_WKSP);

        /* files for writing error */
#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_open_file();
#endif

        /* density */
        double rho = nrhoe*(C_me+C_mp);
        /* electron cyclotron frequency */
        double nu_c = C_e*B/(C_2pi*C_me*C_c);
        /* plasma frequency */
        double nu_p = C_e*sqrt(nrhoe/(C_me*C_pi));
        /* dimensionless electron temperature */
        //double tdl = C_k*Te/C_me_c2;

        fprintf(stdout,"# theta=%.8e  B=%.8e  nrho_e=%.8e  nu_c=%.8e nu_p=%.8e\n",
               theta,B,nrhoe,nu_c,nu_p);
        fflush(stdout);

        /* min and max nu/nu_c */
        double nuratio1=1.e0,nuratio2=1.e4;

        for(i=0;i<npt;i++) {
                /* nu/nu_c */
                double nuratio = nuratio1 * pow(nuratio2/nuratio1,i/((double)(npt-1)));

                double nu = nuratio * nu_c;

                /* emissivity */
                double j = jnu(TOTAL, nu, rho, Te, B, theta, N_WKSP, wksp_g, wksp_n) / nrhoe;

		/* Rybicki & Lightman Eq.(6.33)
		 * j_\nu = P(\omega) / 2   because
                 * j_\nu == dE / (dV d\Omega dt d\nu)
                 *        = (1/4\pi) 2\pi dW / (dV d\Omega)
                 *        = (1/2) dW/(dV d\Omega)
                 *        = P(\omega) / 2 */
		/* xRL is omega / omega_c on R&L p.178
                 * also see Eq.(6.17c) */
		double xRL = nu / ( (3.*GAM0*GAM0*C_e*B*sin(theta))/(C_4pi*C_me*C_c) );
		double epsrel=1.e-10;
		double integralRL,abserr;
		double discard=0.;
		gsl_function func;
		func.function = &funcRL;
		func.params = &discard;
		gsl_integration_qagiu(&func,xRL,0.,epsrel,N_WKSP,wksp_g,&integralRL,&abserr);
		double jRL = ( (sqrt(3.)*C_e3*B*sin(theta))/(C_2pi*C_me_c2) ) * xRL * integralRL / 2.;

		fprintf(stdout,"%.16e %.16e %.16e\n",
		       nuratio, j, jRL
		       );
                fflush(stdout);
        }

        /* free up memory after integration */
        gsl_integration_workspace_free(wksp_g);
        gsl_integration_workspace_free(wksp_n);

#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_close_file() ;
#endif

        return(0) ;
}
