#include "decs.h"

/*============================================================================*/

/* example : isotropic milely relativistic distribution, angle-averaged emissivity
 *
 * Compare j to fitting function in Mahadevan, Narayan & Yi */

/* Choose the temperaure for comparison (see MHY96 Table 1 for details)
 *  FLAG_TE, Temperature in K are
 *  1, 5e8 ; 2, 1e9 ; 3, 2e9 ; 4, 4e9 ; 5, 8e9 ; 6, 1.6e10 ; 7, 3.2e10 */
#define FLAG_TE (7)

/* sanity check */
#if (ELECTRON_DIST != THERMAL)
  #error This file assumes THERMAL distribution is defined.  Set it in decs.h
#endif

/*============================================================================*/

int main(int argc, char *argv[])
{
        double nrhoe = 1.e5;
        double B = 30.;
        /* temperature in K */
        double Te;
        /* observer angle */
        double theta = (45.*C_pi/180.); /* useless in this case */
        /* number of data point, each for one frequency */
        int npt=21;
        int i;

        /* allocation of memory for gamma-n integration: */
        gsl_integration_workspace *wksp_g = gsl_integration_workspace_alloc(N_WKSP);
        gsl_integration_workspace *wksp_n = gsl_integration_workspace_alloc(N_WKSP);

        /* files for writing error */
#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_open_file();
#endif

        /* density */
        double rho = nrhoe*(C_me+C_mp);
        /* electron cyclotron frequency */
        double nu_c = C_e*B/(C_2pi*C_me*C_c);
        /* plasma frequency */
        double nu_p = C_e*sqrt(nrhoe/(C_me*C_pi));

        /* MHY96 */
        double xM,M,R,C;
        double M_alpha,M_beta,M_gamma;
        /* MHY96 Table 1 */
#define FLAG_TE (7)
#if FLAG_TE == 1
        Te = 5.e8; M_alpha = 0.0431; M_beta = 10.44; M_gamma = 16.61;
#elif FLAG_TE == 2
        Te = 1.e9; M_alpha = 1.121; M_beta = -10.65; M_gamma = 9.169;
#elif FLAG_TE == 3
        Te = 2.e9; M_alpha = 1.180; M_beta = -4.008; M_gamma = 1.559;
#elif FLAG_TE == 4
        Te = 4.e9; M_alpha = 1.045; M_beta = -0.1897; M_gamma = 0.0595;
#elif FLAG_TE == 5
        Te = 8.e9; M_alpha = 0.9774; M_beta = 1.160; M_gamma = 0.2641;
#elif FLAG_TE == 6
        Te = 1.6e10; M_alpha = 0.9768; M_beta = 1.095; M_gamma = 0.8332;
#elif FLAG_TE == 7
        Te = 3.2e10; M_alpha = 0.9788; M_beta = 1.021; M_gamma = 1.031;
#else
  #error should not arrive here
#endif
        /* MHY96 Eq.(12) */
        C = C_e2*nrhoe*2.*nu_c/(sqrt(3.)*C_c);

        /* dimensionless electron temperature */
        double tdl = C_k*Te/C_me_c2;

        fprintf(stdout,"# T_e=%.8e  TDL=%.8e  B=%.8e  nrho_e=%.8e  nu_c=%.8e nu_p=%.8e\n",
               Te,tdl,B,nrhoe,nu_c,nu_p);
        fflush(stdout);

        /* min and max nu/nu_c */
        double nuratio1=1.e0,nuratio2=1.e4;

        for(i=0;i<npt;i++) {
                /* nu/nu_c */
                double nuratio = nuratio1 * pow(nuratio2/nuratio1,i/((double)(npt-1)));

                double nu = nuratio * nu_c;

                /* angle-averaged emissivity */
                double j = jnu_average(TOTAL, nu, rho, Te, B, 
                                       N_WKSP, wksp_g, wksp_n);

                /* MHY96 Eq.(12) */
                xM = nuratio/(1.5*tdl*tdl);

                /* MHY96 Eq.(33) */
                M = 4.0505*M_alpha*pow(xM,-1./6)
                    * (1.+0.40*M_beta*pow(xM,-0.25)+0.5316*M_gamma/sqrt(xM))
                    * exp(-1.8899*pow(xM,1./3));

                /* MHY96 Eq.(35) */
                R = nuratio * M * K2_tdlinv_inv_new(1./tdl);

                /* some notes about the notation :
                 * 
                 * According to MNY96 Eq.(34),
                 *   R C = \epsilon_\omega
                 * By MNY96 Eq.(16), there is a factor of 4\pi due to
                 * the pitch-angle dependence.  Therefore,
                 * <j_\nu> = 2 \pi <j_\omega>
                 *         = 2 \pi (\epsilon_\omega / (4 \pi))
                 *         = R C / 2
                 */
                fprintf(stdout,"%.16e %.16e %.16e\n",
                        nuratio, j/nrhoe, R*C/(2.*nrhoe)
                       );
                fflush(stdout);
        }

        /* free up memory after integration */
        gsl_integration_workspace_free(wksp_g);
        gsl_integration_workspace_free(wksp_n);

#if FLAG_ERR1 == ERR1_TO_FILE_CONT || FLAG_ERR2 == ERR2_TO_FILE_CONT
        error_close_file() ;
#endif

        return(0) ;
}
